/*  
Module : NTSERVEVENTLOGSOURCE.CPP
Purpose: Implementation for the class CNTEventLogSource
Created: PJN / 14-07-1997
History: PJN / 23-07-2005 1. Provided an overriden version of Report which has 3 string parameters
                          2. GetStringArrayFromRegistry and SetStringArrayIntoRegistry methods
                          are now public as they are helpful to have available externally
         PJN / 11-01-2006 1. Addition of a default constructor for CNTEventLogSource 
                          2. Addition of CNTEventLogSource::SetServerName, CNTEventLogSource::GetServerName,
                          CNTEventLogSource::SetSourceName and CNTEventLogSource::GetSourceName methods
                          3. Addition of a DWORD* parameter to the CNTEventLogSource::GetStringArrayFromRegistry
                          and CNTEventLogSource::SetStringArrayFromRegistry functions. This allow the details
                          of the last error to be returned to client code.
         PJN / 03-06-2006 1. If CNTEventLogSource::Uninstall fails to remove the HKLM\System\CurrentControlSet\Services\EventLog\"Display Name"
                          registry key, then it aborts before doing any other work. Thanks to Matthias Miller
                          for reporting this issue.
         PJN / 13-06-2006 1. Addition of a logname parameter to the CNTEventLogSource constructor to support user defined event logs
                          instead of logging to the standard "Application" event log. Please note that the code added only enables the CNTService 
                          framework to use non "Application" event logs, but it does not contain code to create a custom event log. You will need 
                          to handle that in your application's install. For further information on doing this, please see 
                          http://msdn.microsoft.com/library/default.asp?url=/library/en-us/eventlog/base/eventlog_key.asp which contains details
                          on the various registry values which you need to create. Thanks to Jan Bartels for this addition.
                          2. Also optimized the construction of member variables in the CNTEventLogSource constructors
         PJN / 24-06-2006 1. Updated code to compile cleanly using VC 2005.
         PJN / 25-06-2006 1. Code now uses newer C++ style casts instead of C style casts.
         PJN / 01-08-2006 1. Fix for two buffer size calculations using the new _tcscpy_s functions, one in CNTEventLogSource::SetStringArrayIntoRegistry and the other in
                          CNTService::GetDependencies. Thanks to DmityShm on the codeproject forums for reporting this issue.
         PJN / 21-09-2006 1. Minor update to use %X instead of %x in various CNTEventLogSource::Report calls.
         PJN / 13-11-2007 1. Minor update to display HRESULT's correctly.

Copyright (c) 1996 - 2008 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


///////////////////////////////// Includes ////////////////////////////////////

#include "stdafx.h"
#include "ntservEventLogSource.h"
#include "ntserv_msg.h"


///////////////////////////////// Macros //////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


///////////////////////////////// Includes ////////////////////////////////////

CNTEventLogSource::CNTEventLogSource() : m_hEventSource(NULL),
                                         m_sLogName(_T("Application"))
{
}

CNTEventLogSource::CNTEventLogSource(LPCWSTR lpUNCServerName, LPCWSTR lpSourceName, LPCWSTR lpLogName) : m_hEventSource(NULL),
                                                                                                         m_sServerName(lpUNCServerName),
                                                                                                         m_sSourceName(lpSourceName),
                                                                                                         m_sLogName(lpLogName)

{
}

CNTEventLogSource::CNTEventLogSource(LPCWSTR lpUNCServerName, LPCWSTR lpSourceName) : m_hEventSource(NULL),
                                                                                      m_sServerName(lpUNCServerName),
                                                                                      m_sSourceName(lpSourceName),
                                                                                      m_sLogName(_T("Application"))
{
}

CNTEventLogSource::~CNTEventLogSource()
{
  Deregister();
}

CNTEventLogSource::operator HANDLE() const
{
  return m_hEventSource;
}

BOOL CNTEventLogSource::Attach(HANDLE hEventSource)
{
  if (m_hEventSource != hEventSource)
    Deregister();

  m_hEventSource = hEventSource;
  return TRUE;
}

HANDLE CNTEventLogSource::Detach()
{
  HANDLE hReturn = m_hEventSource;
  m_hEventSource = NULL;
  return hReturn;
}

BOOL CNTEventLogSource::Register(LPCWSTR lpUNCServerName, LPCWSTR lpSourceName)
{
  Deregister();
  m_hEventSource = ::RegisterEventSource(lpUNCServerName, lpSourceName);
  return (m_hEventSource != NULL);
}

BOOL CNTEventLogSource::Report(WORD wType, WORD wCategory, DWORD dwEventID, PSID lpUserSid,
                               WORD wNumStrings, DWORD dwDataSize, LPCWSTR* lpStrings, LPVOID lpRawData)
{
  CSingleLock sl(&m_csReport, TRUE);

  ASSERT(m_hEventSource == NULL);
  Register(m_sServerName, m_sSourceName);

  //Call the SDK version of the function
  BOOL bSuccess = ::ReportEvent(m_hEventSource, wType,	wCategory, dwEventID, lpUserSid,
                                wNumStrings, dwDataSize, lpStrings, lpRawData);
  Deregister();  

  return bSuccess;
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID)
{
  return Report(wType, 0, dwEventID, NULL, 0, 0, NULL, NULL);
}

BOOL CNTEventLogSource::Report(WORD wType, LPCWSTR lpszString)
{
  return Report(wType, CNTS_MSG_SERVICE_FULLSTRING, lpszString);
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID, LPCWSTR lpszString)
{
  ASSERT(lpszString);
  return Report(wType, 0, dwEventID, NULL, 1, 0, &lpszString, NULL);
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID, LPCWSTR lpszString1, LPCWSTR lpszString2)
{
  ASSERT(lpszString1);
  ASSERT(lpszString2);
  LPCWSTR lpStrings[2];
  lpStrings[0] = lpszString1;
  lpStrings[1] = lpszString2;
  return Report(wType, 0, dwEventID, NULL, 2, 0, lpStrings, NULL);
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID, LPCWSTR lpszString1, LPCWSTR lpszString2, DWORD dwCode, BOOL bReportAsHex)
{
  ASSERT(lpszString1);
  ASSERT(lpszString2);
  LPCWSTR lpStrings[3];
  lpStrings[0] = lpszString1;
  lpStrings[1] = lpszString2;
  CString sError;
  if (bReportAsHex)
    sError.Format(_T("%08X"), dwCode);
  else
    sError.Format(_T("%d"), dwCode);
  lpStrings[2] = sError;
  return Report(wType, 0, dwEventID, NULL, 3, 0, lpStrings, NULL);
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID, LPCWSTR lpszString1, LPCWSTR lpszString2, LPCWSTR lpszString3)
{
  ASSERT(lpszString1);
  ASSERT(lpszString2);
  ASSERT(lpszString3);
  LPCWSTR lpStrings[3];
  lpStrings[0] = lpszString1;
  lpStrings[1] = lpszString2;
  lpStrings[2] = lpszString3;
  return Report(wType, 0, dwEventID, NULL, 3, 0, lpStrings, NULL);
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID, DWORD dwCode, BOOL bReportAsHex)
{
  CString sError;
  if (bReportAsHex)
    sError.Format(_T("%08X"), dwCode);
  else
    sError.Format(_T("%d"), dwCode);
  return Report(wType, dwEventID, sError);
}

BOOL CNTEventLogSource::Report(WORD wType, DWORD dwEventID, LPCWSTR lpszString, DWORD dwCode, BOOL bReportAsHex)
{
  CString sError;
  if (bReportAsHex)
    sError.Format(_T("%08X"), dwCode);
  else
    sError.Format(_T("%d"), dwCode);
  return Report(wType, dwEventID, lpszString, sError);
}

BOOL CNTEventLogSource::Deregister()
{
  BOOL bSuccess = TRUE;
  if (m_hEventSource != NULL)
  {
    bSuccess = ::DeregisterEventSource(m_hEventSource);
    m_hEventSource = NULL;
  }

  return bSuccess;
}

BOOL CNTEventLogSource::Install(LPCWSTR lpSourceName, LPCWSTR lpEventMessageFile, DWORD dwTypesSupported)
{
  //Just delegate the work to the other version of "Install"
  return Install(_T("Application"), lpSourceName, lpEventMessageFile, dwTypesSupported);
}

BOOL CNTEventLogSource::Install(LPCWSTR lpLogName, LPCWSTR lpSourceName, LPCWSTR lpEventMessageFile, DWORD dwTypesSupported)
{
  //Validate our parameters
  AFXASSUME(lpLogName);
  ASSERT(_tcslen(lpLogName));
  AFXASSUME(lpSourceName);
  ASSERT(_tcslen(lpSourceName));
  AFXASSUME(lpEventMessageFile);

  //Make the necessary updates to the registry
  BOOL bSuccess = FALSE;
  HKEY hAppKey;
  CString sKey;
  sKey.Format(_T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\%s"), lpLogName);
  if (::RegOpenKeyEx(HKEY_LOCAL_MACHINE, sKey, 0, KEY_WRITE|KEY_READ, &hAppKey) == ERROR_SUCCESS)
  {
		DWORD dw;
    HKEY hSourceKey;
		if (RegCreateKeyEx(hAppKey, lpSourceName, 0, REG_NONE, REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, NULL,
			                 &hSourceKey, &dw) == ERROR_SUCCESS)
		{
      //Write the Message file string
  		bSuccess = (RegSetValueEx(hSourceKey, _T("EventMessageFile"), NULL, REG_SZ, reinterpret_cast<const BYTE*>(lpEventMessageFile), 
                                (lstrlen(lpEventMessageFile)+1)*sizeof(TCHAR)) == ERROR_SUCCESS);

      //Write the Types supported dword
      bSuccess = bSuccess && (RegSetValueEx(hSourceKey, _T("TypesSupported"), NULL, REG_DWORD,
		                                        reinterpret_cast<const BYTE*>(&dwTypesSupported), sizeof(dwTypesSupported)) == ERROR_SUCCESS);
      //Close the registry key we opened
      ::RegCloseKey(hSourceKey);

      //Update the sources registry key so that the event viewer can filter 
      //on the events which we write to the event log
      CStringArray sources;
      if (GetStringArrayFromRegistry(hAppKey, _T("Sources"), sources))
      {
        //If our name is not in the array then add it
        BOOL bFoundMyself = FALSE;
        for (int i=0; i<sources.GetSize() && !bFoundMyself; i++)
          bFoundMyself = (sources.GetAt(i) == lpSourceName);
        if (!bFoundMyself)
        {
          sources.Add(lpSourceName);
          SetStringArrayIntoRegistry(hAppKey, _T("Sources"), sources);
        }
      }
    }

    //Close the registry key we opened
    ::RegCloseKey(hAppKey);
  }

  return bSuccess;
}

BOOL CNTEventLogSource::Uninstall(LPCWSTR lpSourceName)
{
  //Just delegate the work to the other version of "Install"
  return Uninstall(_T("Application"), lpSourceName);
}

BOOL CNTEventLogSource::Uninstall(LPCWSTR lpLogName, LPCWSTR lpSourceName)
{
  //Validate our parameters
  AFXASSUME(lpLogName);
  ASSERT(_tcslen(lpLogName));
  AFXASSUME(lpSourceName);
  ASSERT(_tcslen(lpSourceName));

  //Remove the settings from the registry
  CString sSubKey;
  sSubKey.Format(_T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\%s\\%s"), lpLogName, lpSourceName);
  long nSuccess = RegDeleteKey(HKEY_LOCAL_MACHINE, sSubKey);
  if (nSuccess != ERROR_SUCCESS) //If we cannot delete this registry key, then abort this function before we go any further
  {
    SetLastError(nSuccess); //Make the last error value available to our callers 
    return FALSE;
  }

  //Remove ourself from the "Sources" registry key
  HKEY hAppKey;
  sSubKey.Format(_T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\%s"), lpLogName);
  if (::RegOpenKeyEx(HKEY_LOCAL_MACHINE, sSubKey, 0, KEY_WRITE | KEY_READ, &hAppKey) == ERROR_SUCCESS)
  {
    CStringArray sources;
    if (GetStringArrayFromRegistry(hAppKey, _T("Sources"), sources))
    {
      //If our name is in the array then remove it
      BOOL bFoundMyself = FALSE;
      for (int i=0; i<sources.GetSize() && !bFoundMyself; i++)
      {
        bFoundMyself = (sources.GetAt(i) == lpSourceName);
        if (bFoundMyself)
          sources.RemoveAt(i);
      }
      if (bFoundMyself)
        SetStringArrayIntoRegistry(hAppKey, _T("Sources"), sources);
    }

    //Close the registry key we opened
    ::RegCloseKey(hAppKey);
  }

  return TRUE;
}

BOOL CNTEventLogSource::GetStringArrayFromRegistry(HKEY hKey, LPCWSTR lpszEntry, CStringArray& array, DWORD* pLastError)
{
	AFXASSUME(hKey != NULL);
	ASSERT(lpszEntry != NULL);

  //Assume the worst
  BOOL bSuccess = FALSE;

  //Empty the array before we go any further
  array.RemoveAll();

	DWORD dwType, dwCount;
	LONG lResult = RegQueryValueEx(hKey, lpszEntry, NULL, &dwType, NULL, &dwCount);
	if (lResult == ERROR_SUCCESS)
	{
		if (dwType == REG_MULTI_SZ)
    {
		  BYTE* lpBuffer = new BYTE[dwCount];
		  lResult = RegQueryValueEx(hKey, lpszEntry, NULL, &dwType, lpBuffer, &dwCount);
      if (lResult == ERROR_SUCCESS)
      {
        LPTSTR lpszStrings = reinterpret_cast<LPTSTR>(lpBuffer);
        while (lpszStrings[0] != 0)
        {
          array.Add(lpszStrings);
          lpszStrings += (_tcslen(lpszStrings ) + 1);
        }

        bSuccess = TRUE;
      }
      else
      {
        if (pLastError)
          *pLastError = lResult;
      }

      //Tidy up the heap memory we have used
      delete [] lpBuffer;
    }
    else
    {
      SetLastError(ERROR_INVALID_DATA);

      if (pLastError)
        *pLastError = ERROR_INVALID_DATA;
    }
	}
  else
  {
    if (pLastError)
      *pLastError = lResult;
  }

  return bSuccess;
}

BOOL CNTEventLogSource::SetStringArrayIntoRegistry(HKEY hKey, LPCWSTR lpszEntry, const CStringArray& array, DWORD* pLastError)
{   
  //Validate our input parameters
  AFXASSUME(hKey);
  ASSERT(lpszEntry != NULL);

  //Work out the size of the buffer we will need
  DWORD dwSize = 0;
  INT_PTR nStrings = array.GetSize();
  for (INT_PTR i=0; i<nStrings; i++)
    dwSize += array.GetAt(i).GetLength() + 1; //1 extra for each NULL terminator

  //Need one second NULL for the double NULL at the end
  dwSize++;

  //Convert from BYTE size to character size
  dwSize *= sizeof(TCHAR);

  //Allocate the memory we want
  BYTE* lpBuffer = new BYTE[dwSize];
  memset(lpBuffer, 0, dwSize);

  //Now copy the strings into the buffer
  int nCurOffset = 0;
  LPTSTR lpszString = reinterpret_cast<LPTSTR>(lpBuffer);
  for (INT_PTR i=0; i<nStrings; i++)
  {
    CString sText = array.GetAt(i);
    int nCurrentStringLength = sText.GetLength();
    _tcscpy_s(&lpszString[nCurOffset], nCurrentStringLength+1, sText);
    nCurOffset += (nCurrentStringLength + 1);
  }
  //Don't forgot to doubly NULL terminate
  lpszString[nCurOffset] = _T('\0');

  //Finally write it into the registry
  LONG lResult = ::RegSetValueEx(hKey, lpszEntry, NULL, REG_MULTI_SZ, lpBuffer, dwSize);
	BOOL bSuccess = (lResult == ERROR_SUCCESS);
  if (!bSuccess && pLastError)
    *pLastError = lResult;

  //free up the memory we used
  delete [] lpBuffer;

  return bSuccess;
}
